/*
  Exercice 13

  Effectuez des diffusions sur les lignes de processus et des sommes globales sur
  les colonnes. 
*/

#include <math.h>
#include <mpi.h>

/*#define DEBUG*/

void rotate(MPI_Comm, int);

void main(argv, argc)
     int argv; 
     char *argc[];
     
{
  int size,          /* nombre de processus                         */
    my_num_p,        /* mon numero avant                            */
    my_num_a,        /* mon numero apres la creation de la grille   */
    dims[2],         /* dimensions et nombre de procs ds chaque dim */
    periods[2],      /* periodes pour chaque dimension              */
    mycoords[2],     /* coordonnees dans la grille virtuelle        */
    remain_dims[2],  /* tableau des dimensions restantes (ss-grille)*/
    sqrp,            /* racine carre du nombre de proc              */
    reorder;         /* re-arrangement des processus                */
  MPI_Comm new_comm, /* communicateur de la topologie cartesienne   */
    col_comm,        /* communicateur colonnes                      */
    row_comm;        /* communicateur lignes                        */

 
  MPI_Init(&argv, &argc);
  
  MPI_Comm_rank(MPI_COMM_WORLD, &my_num_p);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  sqrp = (int)(sqrt((double)size));
  if (my_num_p == 0)
    printf("\nCreation d'une grille %dx%d\n\n", sqrp, sqrp);

  /*
  if (sqrp*sqrp != size) {
    printf("%d pas un nombre carre de procs, desole !\n", size);
    exit(1);
  }*/
  
  printf("Hello world from %d out of %d\n", my_num_p, size);
  
  /* 
   * Creation de la topologie cartesienne 
   */
  dims[0] = dims[1] = sqrp;  /* grille carre de sqrp*sqrp          */
  periods[0] = 1;            /* periode sur les colonnes           */
  periods[1] = 1;            /* pas de periode sur les lignes      */
  reorder = 1;               /* on re-arrange les numeros de procs */
  MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder, &new_comm);

  /*
   * recuperation des coordonnees dans le nouveau communicateur
   */
  if (new_comm != MPI_COMM_NULL) {
    MPI_Comm_rank(new_comm, &my_num_a);
    MPI_Cart_coords(new_comm, my_num_a, 2, mycoords);
    printf("Hello de %d (avt %d), coords (%d,%d)\n", my_num_a, my_num_p, 
	   mycoords[0], mycoords[1]);

    /* 
     * decoupage de la grille en lignes
     */
    remain_dims[0] = 0; 
    remain_dims[1] = 1;
    MPI_Cart_sub(new_comm, remain_dims, &row_comm);

    /* 
     * decoupage de la grille en colonnes 
     */
    remain_dims[0] = 1; 
    remain_dims[1] = 0;
    MPI_Cart_sub(new_comm, remain_dims, &col_comm);
    
    /*
     * rotation jeton ligne
     */
    rotate(row_comm, 12);

     /*
     * rotation jeton colonne
     */
    rotate(col_comm, 22);
  }
  else {
    printf("%d n'existe plus dans la nouvelle grille\n", my_num_p);
  }

  MPI_Finalize();
}

/*
 * Rotation d'un jeton a partir du processeur 0 du communicateur donne
 */
void rotate(MPI_Comm comm, int jet)
{
  int size, right, left, jeton, me;
  MPI_Status status;

  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &me);
 
  left = (me - 1 + size)%size;
  right = (me + 1)%size;

  if (me == 0) {  
    jeton = jet;
    MPI_Send(&jeton, 1, MPI_INT, right, 1, comm);
#ifdef DEBUG
    printf("%d a envoye %d a %d\n", me, jeton, right);
#endif    

    MPI_Recv(&jeton, 1, MPI_INT, left, 1, comm, &status);
    /*#ifdef DEBUG*/
    printf("%d a recu %d de %d\n", me, jeton, left);
    /*#endif*/
  }
  else {
    MPI_Recv(&jeton, 1, MPI_INT, left, 1, comm, &status);
#ifdef DEBUG
    printf("%d a recu %d de %d\n", me, jeton, left);
#endif    
    
    jeton++; /* modif jeton */
    
    MPI_Send(&jeton, 1, MPI_INT, right, 1, comm);
#ifdef DEBUG
    printf("%d a envoye %d a %d\n", me, jeton, right);
#endif    
  }
}
