/*
  Exercice 14

  Echangez les bords d'une matrice avec vos voisins dans la grille.


    ------------------------------
    |  |                      |  |
    ------------------------------
    |  |                      |  |
    |  |                      |  |
    |  |                      |  |
    |  |                      |  |
    |  |                      |  |
    ------------------------------
    |  |                      |  |
    ------------------------------

*/

#include <math.h>
#include <mpi.h>

#define DEBUG

#define NUM_VERIF 4

#define TAILLE_MAT 4

void main(argv, argc)
     int argv; 
     char *argc[];
{
  int size,          /* nombre de processus                         */
    sizef,           /* taille de la grille carre                   */
    my_num_p,        /* mon numero avant                            */
    my_num_a,        /* mon numero apres la creation de la grille   */
    dims[2],         /* dimensions et nombre de procs ds chaque dim */
    periods[2],      /* periodes pour chaque dimension              */
    mycoords[2],     /* coordonnees dans la grille virtuelle        */
    hiscoords[2],    /* temporaire                                  */
    haut, bas, gauche, droite,
    remain_dims[2],  /* tableau des dimensions restantes (ss-grille)*/
    sqrp,            /* racine carre du nombre de proc              */
    reorder,         /* re-arrangement des processus                */
    tab[TAILLE_MAT][TAILLE_MAT],    
                     /* matrice utilisee pour les communications (E)*/
    tab1[TAILLE_MAT][TAILLE_MAT],    
                     /* matrice utilisee pour les communications (R)*/
    i, j, k, dem;
  MPI_Comm new_comm, /* communicateur de la topologie cartesienne   */
    col_comm,        /* communicateur colonnes                      */
    row_comm;        /* communicateur lignes                        */
  MPI_Datatype col_type, /* type de tableau colonne                 */
    row_type;        /* type de tableau ligne                       */
  MPI_Request request[4]; /* tableau pour les coms non-bloquantes   */
  MPI_Status status[4]; /* tableau de status de reception           */
 
  MPI_Init(&argv, &argc);
  
  MPI_Comm_rank(MPI_COMM_WORLD, &my_num_p);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  sqrp = (int)(sqrt((double)size));
  if (my_num_p == 0)
    printf("\nCreation d'une grille %dx%d\n\n", sqrp, sqrp);

  printf("Hello world from %d out of %d\n", my_num_p, size);
  
  /* 
   * Creation de la topologie cartesienne 
   */
  dims[0] = dims[1] = sqrp;  /* grille carre de sqrp*sqrp          */
  periods[0] = 1;            /* periode sur les colonnes           */
  periods[1] = 1;            /* pas de periode sur les lignes      */
  reorder = 1;               /* on re-arrange les numeros de procs */
  MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder, &new_comm);

  /*
   * recuperation des coordonnees dans le nouveau communicateur
   */
  if (new_comm != MPI_COMM_NULL) {
    MPI_Comm_rank(new_comm, &my_num_a);
    MPI_Cart_coords(new_comm, my_num_a, 2, mycoords);
    MPI_Comm_size(new_comm, &sizef);

#ifdef DEBUG
    printf("Hello de %d (avt %d), coords (%d,%d)\n", my_num_a, my_num_p, 
	   mycoords[0], mycoords[1]);
#endif

    /* Initialisation des matrices locales */  
    dem = mycoords[0]*TAILLE_MAT*TAILLE_MAT*sqrp + 
      mycoords[1]*TAILLE_MAT;

#ifdef DEBUG
    printf("Sur %d, dem=%d\n", my_num_a, dem);
#endif

    for (i=0; i<TAILLE_MAT; i++) {
      for (j=0; j<TAILLE_MAT; j++) {
	tab[i][j] = dem++;
	tab1[i][j] = 0;
      }
      dem += TAILLE_MAT;
    }

    /* verif matrice locale */
#ifdef DEBUG
    for (k=0; k<sizef; k++) {
      MPI_Barrier(new_comm);

      if (my_num_a == k) {
	printf("\nCool, %d affiche sa matrice !\n\n", my_num_a);
	for (i=0; i<TAILLE_MAT; i++) {
	  for (j=0; j<TAILLE_MAT; j++) {
	    printf("%d\t", tab[i][j]);
	  }
	  printf("\n");
	}
      }
    }  
    MPI_Barrier(new_comm);
#endif  

    /* creation du type ligne */
    MPI_Type_contiguous(TAILLE_MAT, MPI_INT, &row_type);
    MPI_Type_commit(&row_type);

    /* creation du type "colonne" */
    MPI_Type_vector(TAILLE_MAT, 1, TAILLE_MAT, MPI_INT, &col_type);
    MPI_Type_commit(&col_type);

    /* calcul des voisins ds la grille */
    hiscoords[0] = (mycoords[0] - 1)%sqrp;
    hiscoords[1] = mycoords[1];
    MPI_Cart_rank(new_comm, hiscoords, &haut);

    hiscoords[0] = (mycoords[0] + 1)%sqrp;
    MPI_Cart_rank(new_comm, hiscoords, &bas);

    hiscoords[0] = mycoords[0];
    hiscoords[1] = (mycoords[1]+1)%sqrp;
    MPI_Cart_rank(new_comm, hiscoords, &droite);

    hiscoords[1] = (mycoords[1]-1)%sqrp;
    MPI_Cart_rank(new_comm, hiscoords, &gauche);

#ifdef DEBUG
    printf("Pour %d, envoi (h,b,d,g)=(%d,%d,%d,%d)\n",
	   my_num_a, haut, bas, droite, gauche);
#endif

    /* Envoi non-bloquant des bords de matrices */
    MPI_Isend(&(tab[0][0]), 1, row_type, haut, 0, 
	      new_comm, &(request[0]));
    MPI_Isend(&(tab[TAILLE_MAT-1][0]), 1, row_type, bas, 0, 
	      new_comm, &(request[1]));
    MPI_Isend(&(tab[0][0]), 1, col_type, gauche, 0, 
	      new_comm, &(request[2]));
    MPI_Isend(&(tab[0][TAILLE_MAT-1]), 1, col_type, droite, 0, 
	      new_comm, &(request[3]));
#ifdef DEBUG
    printf("Sur %d, emissions non-bloquantes effectuees\n", my_num_a);
#endif

    /* Reception bloquante des bords */
    MPI_Recv(&(tab1[0][0]), 1, row_type, haut, 0, 
	      new_comm, status);
    MPI_Recv(&(tab1[TAILLE_MAT-1][0]), 1, row_type, bas, 0, 
	      new_comm, status);
    MPI_Recv(&(tab1[0][0]), 1, col_type, gauche, 0, 
	      new_comm, status);
    MPI_Recv(&(tab1[0][TAILLE_MAT-1]), 1, col_type, droite, 0, 
	      new_comm, status);

#ifdef DEBUG
    printf("Sur %d, reception effectuees\n", my_num_a);
#endif

    /* Verification de la fin des envois */
    MPI_Waitall(4, request, status);
#ifdef DEBUG
    printf("Sur %d, emissions terminees\n", my_num_a);
#endif

    /* verif resultat */
    MPI_Barrier(new_comm); /* pour ne pas melanger les affichages ! */

    if (my_num_a == NUM_VERIF) {
      printf("\nCool, %d affiche le resultat !\n\n", my_num_a);
      for (i=0; i<TAILLE_MAT; i++) {
	for (j=0; j<TAILLE_MAT; j++) {
	  printf("%d\t", tab1[i][j]);
	}
	printf("\n");
      }
    }
  }
  else {
    printf("%d n'existe plus dans la nouvelle grille\n", my_num_p);
  }

  MPI_Finalize();
}

